unit ExtBtn;

interface
uses Windows, Messages, Classes, Controls, Forms, Graphics, StdCtrls,
  ExtCtrls, CommCtrl;

type
  TButtonLayout = (blGlyphLeft, blGlyphRight, blGlyphTop, blGlyphBottom);
  TButtonState  = (bsUp, bsDisabled, bsDown, bsExclusive);
  TButtonStyle  = (bsAutoDetect, bsWin31, bsNew);
  TKind         = (knClose, knMinimize, knNone);
  TNumGlyphs = 1..4;

  TExtBtn = class(TGraphicControl)
  private
    FGroupIndex: Integer;
    FGlyph: Pointer;
    FExtGlyph: Pointer;
    FDown: Boolean;
    FDragging: Boolean;
    FAllowAllUp: Boolean;
    FLayout: TButtonLayout;
    FSpacing: Integer;
    FTransparent: Boolean;
    FMargin: Integer;
    FExtMargin: Integer;
    FFlat: Boolean;
    FMouseInControl: Boolean;
    FExtButton : boolean;
    FExtWidth : integer;
    FExtTransparent : boolean;
    FOnExtBtnClick : TNotifyEvent;
    FKind: TKind;
    FBevelShow: Boolean;
    FThinBorder: Boolean;
    FHotTrack: Boolean;
    FNormalColor: TColor;
    FHotColor: TColor;
    FDownColor: TColor;
    FBtnColor: TColor;
    FNeedManualMouseUp: Boolean;
    procedure GlyphChanged(Sender: TObject);
    procedure UpdateExclusive;
    function  GetGlyph: TBitmap;
    procedure SetGlyph(Value: TBitmap);
    function  GetNumGlyphs: TNumGlyphs;
    procedure SetNumGlyphs(Value: TNumGlyphs);
    procedure SetDown(Value: Boolean);
    procedure SetHotTrack(Value: Boolean);
    procedure SetFlat(Value: Boolean);
    procedure SetAllowAllUp(Value: Boolean);
    procedure SetGroupIndex(Value: Integer);
    procedure SetLayout(Value: TButtonLayout);
    procedure SetSpacing(Value: Integer);
    procedure SetTransparent(Value: Boolean);
    procedure SetThinBorder(Value: Boolean);
    procedure SetMargin(Value: Integer);
    procedure SetExtMargin(Value: Integer);
    procedure SetExtButton(Value: Boolean);
    procedure SetBevelShow(Value: Boolean);
    procedure SetExtWidth(Value: integer);
    procedure SetExtTransparent(Value: Boolean);
    procedure SetKind(Value: TKind);
	procedure SetBtnColor(Value: TColor);
	procedure SetNormalColor(Value: TColor);
	procedure SetDownColor(Value: TColor);
    procedure WMLButtonDblClk(var Message: TWMLButtonDown); message WM_LBUTTONDBLCLK;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CMButtonPressed(var Message: TMessage); message CM_BUTTONPRESSED;
    procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
    procedure CMSysColorChange(var Message: TMessage); message CM_SYSCOLORCHANGE;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
  protected
    FState: TButtonState;
    FExtState: TButtonState;
    procedure ActionChange(Sender: TObject; CheckDefaults: Boolean); override;
    function GetPalette: HPALETTE; override;
    procedure Loaded; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;  X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;    X, Y: Integer); override;
    procedure Paint; override;
    property MouseInControl: Boolean read FMouseInControl;
    procedure UpdateTracking;
  public
	procedure MouseManualUp();
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Click; override;
    procedure UpdateMouseInControl;
  published
    property Align default alLeft;
    property Action;
    property AllowAllUp: Boolean read FAllowAllUp write SetAllowAllUp default False;
    property Anchors;
    property BiDiMode;
    property Constraints;
    property BevelShow: boolean read FBevelShow write SetBevelShow default true;
    property HotTrack: boolean read FHotTrack write SetHotTrack default false;
    property HotColor: TColor read FHotColor write FHotColor default clBlue;
    property NormalColor: TColor read FNormalColor write SetNormalColor default clBlack;
    property BtnColor: TColor read FBtnColor write SetBtnColor default clBtnFace;
    property DownColor: TColor read FDownColor write SetDownColor default clBtnHighlight;
    property ExtButton: boolean read FExtButton write SetExtButton default false;
    property ExtWidth: integer read FExtWidth write SetExtWidth default 14;
    property ExtTransparent: boolean read FExtTransparent write SetExtTransparent default false;
    property GroupIndex: Integer read FGroupIndex write SetGroupIndex default 0;
    property Down: Boolean read FDown write SetDown default False;
    property Caption;
    property Enabled;
    property Flat: Boolean read FFlat write SetFlat default True;
    property Font;
    property Glyph: TBitmap read GetGlyph write SetGlyph;
    property Kind: TKind read FKind write SetKind default knClose;
    property Layout: TButtonLayout read FLayout write SetLayout default blGlyphLeft;
    property Margin: Integer read FMargin write SetMargin default -1;
    property ExtMargin: Integer read FExtMargin write SetExtMargin default -1;
    property NumGlyphs: TNumGlyphs read GetNumGlyphs write SetNumGlyphs default 1;
    property ParentFont;
    property ParentShowHint;
    property ParentBiDiMode;
    property PopupMenu;
    property ShowHint;
    property Spacing: Integer read FSpacing write SetSpacing default 4;
    property Transparent: Boolean read FTransparent write SetTransparent default True;
    property FlatAlwaysEdge: Boolean read FThinBorder write SetThinBorder default False;
    property Visible;
    property OnExtBtnClick : TNotifyEvent read FOnExtBtnClick write FOnExtBtnClick;
    property OnClick;
    property OnDblClick;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnDragDrop;
    property OnDragOver;
    property OnStartDrag;
  end;

const
    ExtButtonList: TList = nil;

implementation
uses Registry, Consts, ActnList, ImgList, SysUtils;
{$R *.Res}

type
  TGlyphList = class(TImageList)
  private
    Used: TBits;
    FCount: Integer;
    function AllocateIndex: Integer;
  public
    constructor CreateSize(AWidth, AHeight: Integer);
    destructor Destroy; override;
    function AddMasked(Image: TBitmap; MaskColor: TColor): Integer;
    procedure Delete(Index: Integer);
    property Count: Integer read FCount;
  end;

  TGlyphCache = class
  private
    GlyphLists: TList;
  public
    constructor Create;
    destructor Destroy; override;
    function GetList(AWidth, AHeight: Integer): TGlyphList;
    procedure ReturnList(List: TGlyphList);
    function Empty: Boolean;
  end;

  TButtonGlyph = class
  private
  	FBtnColor: TColor;
    FOriginal: TBitmap;
    FGlyphList: TGlyphList;
    FIndexs: array[TButtonState] of Integer;
    FTransparentColor: TColor;
    FNumGlyphs: TNumGlyphs;
    FOnChange: TNotifyEvent;
    procedure GlyphChanged(Sender: TObject);
    procedure SetGlyph(Value: TBitmap);
    procedure SetNumGlyphs(Value: TNumGlyphs);
    procedure Invalidate;
    function CreateButtonGlyph(State: TButtonState): Integer;
    procedure DrawButtonGlyph(Canvas: TCanvas; const GlyphPos: TPoint;
      State: TButtonState; Transparent: Boolean);
    procedure DrawButtonText(Canvas: TCanvas; const Caption: string;
      TextBounds: TRect; State: TButtonState; BiDiFlags: Longint);
    procedure CalcButtonLayout(Canvas: TCanvas; const Client: TRect;
      const Offset: TPoint; const Caption: string; Layout: TButtonLayout;
      Margin, Spacing: Integer; var GlyphPos: TPoint; var TextBounds: TRect;
      BiDiFlags: Longint);
  public
    constructor Create;
    destructor Destroy; override;
    { return the text rectangle }
    function Draw(Canvas: TCanvas; const Client: TRect; const Offset: TPoint;
      const Caption: string; Layout: TButtonLayout; Margin, Spacing: Integer;
      State: TButtonState; Transparent: Boolean; BiDiFlags: Longint): TRect;
    property Glyph: TBitmap read FOriginal write SetGlyph;
    property NumGlyphs: TNumGlyphs read FNumGlyphs write SetNumGlyphs;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

{ TGlyphList }

constructor TGlyphList.CreateSize(AWidth, AHeight: Integer);
begin
  inherited CreateSize(AWidth, AHeight);
  Used := TBits.Create;
end;

destructor TGlyphList.Destroy;
begin
  Used.Free;
  inherited Destroy;
end;
    
function TGlyphList.AllocateIndex: Integer;
begin
  Result := Used.OpenBit;
  if Result >= Used.Size then
  begin
    Result := inherited Add(nil, nil);
    Used.Size := Result + 1;
  end;
  Used[Result] := True;
end;
    
function TGlyphList.AddMasked(Image: TBitmap; MaskColor: TColor): Integer;
begin
  Result := AllocateIndex;
  ReplaceMasked(Result, Image, MaskColor);
  Inc(FCount);
end;
    
procedure TGlyphList.Delete(Index: Integer);
begin
  if Used[Index] then
  begin
    Dec(FCount);
    Used[Index] := False;
  end;
end;
    
{ TGlyphCache }
    
constructor TGlyphCache.Create;
begin
  inherited Create;
  GlyphLists := TList.Create;
end;

destructor TGlyphCache.Destroy;
begin
  GlyphLists.Free;
  inherited Destroy;
end;
    
function TGlyphCache.GetList(AWidth, AHeight: Integer): TGlyphList;
var
  I: Integer;
begin
  for I := GlyphLists.Count - 1 downto 0 do
  begin
    Result := GlyphLists[I];
    with Result do
      if (AWidth = Width) and (AHeight = Height) then Exit;
  end;
  Result := TGlyphList.CreateSize(AWidth, AHeight);
  GlyphLists.Add(Result);
end;
    
procedure TGlyphCache.ReturnList(List: TGlyphList);
begin
  if List = nil then Exit;
  if List.Count = 0 then
  begin
    GlyphLists.Remove(List);
    List.Free;
  end;
end;
    
function TGlyphCache.Empty: Boolean;
begin
  Result := GlyphLists.Count = 0;
end;
    
var
  GlyphCache: TGlyphCache = nil;
  ButtonCount: Integer = 0;

{ TButtonGlyph }

constructor TButtonGlyph.Create;
var
  I: TButtonState;
begin
  inherited Create;
  FOriginal := TBitmap.Create;
  FOriginal.OnChange := GlyphChanged;
  FTransparentColor := clWhite;
  FNumGlyphs := 1;
  for I := Low(I) to High(I) do
    FIndexs[I] := -1;
  if GlyphCache = nil then GlyphCache := TGlyphCache.Create;
end;

destructor TButtonGlyph.Destroy;
begin
  FOriginal.Free;
  Invalidate;
  if Assigned(GlyphCache) and GlyphCache.Empty then
  begin
    GlyphCache.Free;
    GlyphCache := nil;
  end;
  inherited Destroy;
end;

procedure TButtonGlyph.Invalidate;
var
  I: TButtonState;
begin
  for I := Low(I) to High(I) do
  begin
    if FIndexs[I] <> -1 then FGlyphList.Delete(FIndexs[I]);
    FIndexs[I] := -1;
  end;
  GlyphCache.ReturnList(FGlyphList);
  FGlyphList := nil;
end;

procedure TButtonGlyph.GlyphChanged(Sender: TObject);
begin
  if Sender = FOriginal then
  begin
    FTransparentColor := clWhite;
//    FOriginal.TransparentColor;
//------------------------------------------------------------------
    Invalidate;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TButtonGlyph.SetGlyph(Value: TBitmap);
var
  Glyphs: Integer;
begin
  Invalidate;
  FOriginal.Assign(Value);
  if (Value <> nil) and (Value.Height > 0) then
  begin
    FTransparentColor := clWhite;
//Value.TransparentColor;
//-------------------------------------------------------------------
    if Value.Width mod Value.Height = 0 then
    begin
      Glyphs := Value.Width div Value.Height;
      if Glyphs > 4 then Glyphs := 1;
      SetNumGlyphs(Glyphs);
    end;
  end;
end;
    
procedure TButtonGlyph.SetNumGlyphs(Value: TNumGlyphs);
begin
  if (Value <> FNumGlyphs) and (Value > 0) then
  begin
    Invalidate;
    FNumGlyphs := Value;
    GlyphChanged(Glyph);
  end;
end;
    
function TButtonGlyph.CreateButtonGlyph(State: TButtonState): Integer;
const
  ROP_DSPDxax = $00E20746;
var
  TmpImage, DDB, MonoBmp: TBitmap;
  IWidth, IHeight: Integer;
  IRect, ORect: TRect;
  I: TButtonState;
  DestDC: HDC;
begin
  if (State = bsDown) and (NumGlyphs < 3) then State := bsUp;
  Result := FIndexs[State];
  if Result <> -1 then Exit;
  if (FOriginal.Width or FOriginal.Height) = 0 then Exit;
  IWidth := FOriginal.Width div FNumGlyphs;
  IHeight := FOriginal.Height;
  if FGlyphList = nil then
  begin
    if GlyphCache = nil then GlyphCache := TGlyphCache.Create;
    FGlyphList := GlyphCache.GetList(IWidth, IHeight);
  end;
  TmpImage := TBitmap.Create;
  try
    TmpImage.Width := IWidth;
    TmpImage.Height := IHeight;
    IRect := Rect(0, 0, IWidth, IHeight);
    TmpImage.Canvas.Brush.Color := FBtnColor;
    TmpImage.Palette := CopyPalette(FOriginal.Palette);
    I := State;
    if Ord(I) >= NumGlyphs then I := bsUp;
    ORect := Rect(Ord(I) * IWidth, 0, (Ord(I) + 1) * IWidth, IHeight);
    case State of
      bsUp, bsDown,
      bsExclusive:
        begin
          TmpImage.Canvas.CopyRect(IRect, FOriginal.Canvas, ORect);
          if FOriginal.TransparentMode = tmFixed then
            FIndexs[State] := FGlyphList.AddMasked(TmpImage, FTransparentColor)
          else
            FIndexs[State] := FGlyphList.AddMasked(TmpImage, clDefault);
        end;
      bsDisabled:
        begin
          MonoBmp := nil;
          DDB := nil;
          try
            MonoBmp := TBitmap.Create;
            DDB := TBitmap.Create;
            DDB.Assign(FOriginal);
            DDB.HandleType := bmDDB;
            if NumGlyphs > 1 then
            with TmpImage.Canvas do
            begin    { Change white & gray to clBtnHighlight and clBtnShadow }
              CopyRect(IRect, DDB.Canvas, ORect);
              MonoBmp.Monochrome := True;
              MonoBmp.Width := IWidth;
              MonoBmp.Height := IHeight;
    
              { Convert white to clBtnHighlight }
              DDB.Canvas.Brush.Color := clWhite;
              MonoBmp.Canvas.CopyRect(IRect, DDB.Canvas, ORect);
              Brush.Color := clBtnHighlight;
              DestDC := Handle;
              SetTextColor(DestDC, clBlack);
              SetBkColor(DestDC, clWhite);
              BitBlt(DestDC, 0, 0, IWidth, IHeight,
                     MonoBmp.Canvas.Handle, 0, 0, ROP_DSPDxax);
    
              { Convert gray to clBtnShadow }
              DDB.Canvas.Brush.Color := clGray;
              MonoBmp.Canvas.CopyRect(IRect, DDB.Canvas, ORect);
              Brush.Color := clBtnShadow;
              DestDC := Handle;
              SetTextColor(DestDC, clBlack);
              SetBkColor(DestDC, clWhite);
              BitBlt(DestDC, 0, 0, IWidth, IHeight,
                     MonoBmp.Canvas.Handle, 0, 0, ROP_DSPDxax);

              { Convert transparent color to clBtnFace }
              DDB.Canvas.Brush.Color := ColorToRGB(FTransparentColor);
              MonoBmp.Canvas.CopyRect(IRect, DDB.Canvas, ORect);
              Brush.Color := FBtnColor;
              DestDC := Handle;
              SetTextColor(DestDC, clBlack);
              SetBkColor(DestDC, clWhite);
              BitBlt(DestDC, 0, 0, IWidth, IHeight,
                     MonoBmp.Canvas.Handle, 0, 0, ROP_DSPDxax);
            end
            else
            begin
              { Create a disabled version }
              with MonoBmp do
              begin
                Assign(FOriginal);
                HandleType := bmDDB;
                Canvas.Brush.Color := clBlack;
                Width := IWidth;
                if Monochrome then
                begin
                  Canvas.Font.Color := clWhite;
                  Monochrome := False;
                  Canvas.Brush.Color := clWhite;
                end;
                Monochrome := True;
              end;
              with TmpImage.Canvas do
              begin
                Brush.Color := FBtnColor;
                FillRect(IRect);
                Brush.Color := clBtnHighlight;
                SetTextColor(Handle, clBlack);
                SetBkColor(Handle, clWhite);
                BitBlt(Handle, 1, 1, IWidth, IHeight,
                  MonoBmp.Canvas.Handle, 0, 0, ROP_DSPDxax);
                Brush.Color := clBtnShadow;
                SetTextColor(Handle, clBlack);
                SetBkColor(Handle, clWhite);
                BitBlt(Handle, 0, 0, IWidth, IHeight,
                  MonoBmp.Canvas.Handle, 0, 0, ROP_DSPDxax);
              end;
            end;
          finally
            DDB.Free;
            MonoBmp.Free;
          end;
          FIndexs[State] := FGlyphList.AddMasked(TmpImage, clDefault);
        end;
    end;
  finally
    TmpImage.Free;
  end;
  Result := FIndexs[State];
  FOriginal.Dormant;
end;
    
procedure TButtonGlyph.DrawButtonGlyph(Canvas: TCanvas; const GlyphPos: TPoint;
  State: TButtonState; Transparent: Boolean);
var
  Index: Integer;
begin
  if FOriginal = nil then Exit;
  if (FOriginal.Width = 0) or (FOriginal.Height = 0) then Exit;
  Index := CreateButtonGlyph(State);
  with GlyphPos do
    if Transparent or (State = bsExclusive) then
      ImageList_DrawEx(FGlyphList.Handle, Index, Canvas.Handle, X, Y, 0, 0,
        clNone, clNone, ILD_Transparent)
    else
      ImageList_DrawEx(FGlyphList.Handle, Index, Canvas.Handle, X, Y, 0, 0,
        ColorToRGB(FBtnColor), clNone, ILD_Normal);   
end;

procedure TButtonGlyph.DrawButtonText(Canvas: TCanvas; const Caption: string;
  TextBounds: TRect; State: TButtonState; BiDiFlags: LongInt);
begin
  with Canvas do
  begin
    Brush.Style := bsClear;
//.	Font.Color  := FSaveColor;
    if State = bsDisabled then
    begin
      OffsetRect(TextBounds, 1, 1);
      Font.Color := clBtnHighlight;
      DrawText(Handle, PChar(Caption), Length(Caption), TextBounds,
        DT_CENTER or DT_VCENTER or BiDiFlags);
      OffsetRect(TextBounds, -1, -1);
      Font.Color := clBtnShadow;
      DrawText(Handle, PChar(Caption), Length(Caption), TextBounds,
        DT_CENTER or DT_VCENTER or BiDiFlags);
    end else
      DrawText(Handle, PChar(Caption), Length(Caption), TextBounds,
        DT_CENTER or DT_VCENTER or BiDiFlags);
  end;
end;
    
procedure TButtonGlyph.CalcButtonLayout(Canvas: TCanvas; const Client: TRect;
  const Offset: TPoint; const Caption: string; Layout: TButtonLayout; Margin,
  Spacing: Integer; var GlyphPos: TPoint; var TextBounds: TRect;
  BiDiFlags: LongInt);
var
  TextPos: TPoint;
  ClientSize, GlyphSize, TextSize: TPoint;
  TotalSize: TPoint;
begin
  if (BiDiFlags and DT_RIGHT) = DT_RIGHT then
    if Layout = blGlyphLeft then Layout := blGlyphRight
    else
      if Layout = blGlyphRight then Layout := blGlyphLeft;
  { calculate the item sizes }
  ClientSize := Point(Client.Right - Client.Left, Client.Bottom -
    Client.Top);
    
  if FOriginal <> nil then
    GlyphSize := Point(FOriginal.Width div FNumGlyphs, FOriginal.Height) else
    GlyphSize := Point(0, 0);
    
  if Length(Caption) > 0 then
  begin
    TextBounds := Rect(0, 0, Client.Right - Client.Left, 0);
    DrawText(Canvas.Handle, PChar(Caption), Length(Caption), TextBounds,
      DT_CALCRECT or BiDiFlags);
    TextSize := Point(TextBounds.Right - TextBounds.Left, TextBounds.Bottom -
      TextBounds.Top);
  end
  else
  begin
    TextBounds := Rect(0, 0, 0, 0);
    TextSize := Point(0,0);
  end;

  { If the layout has the glyph on the right or the left, then both the
    text and the glyph are centered vertically.  If the glyph is on the top
    or the bottom, then both the text and the glyph are centered horizontally.}
  if Layout in [blGlyphLeft, blGlyphRight] then
  begin
    GlyphPos.Y := (ClientSize.Y - GlyphSize.Y + 1) div 2;
    TextPos.Y := (ClientSize.Y - TextSize.Y + 1) div 2;
  end
  else
  begin
    GlyphPos.X := (ClientSize.X - GlyphSize.X + 1) div 2;
    TextPos.X := (ClientSize.X - TextSize.X + 1) div 2;
  end;

  { if there is no text or no bitmap, then Spacing is irrelevant }
  if (TextSize.X = 0) or (GlyphSize.X = 0) then
    Spacing := 0;
    
  { adjust Margin and Spacing }
  if Margin = -1 then
  begin
    if Spacing = -1 then
    begin
      TotalSize := Point(GlyphSize.X + TextSize.X, GlyphSize.Y + TextSize.Y);
      if Layout in [blGlyphLeft, blGlyphRight] then
        Margin := (ClientSize.X - TotalSize.X) div 3
      else
        Margin := (ClientSize.Y - TotalSize.Y) div 3;
      Spacing := Margin;
    end
    else
    begin
      TotalSize := Point(GlyphSize.X + Spacing + TextSize.X, GlyphSize.Y +
        Spacing + TextSize.Y);
      if Layout in [blGlyphLeft, blGlyphRight] then
        Margin := (ClientSize.X - TotalSize.X + 1) div 2
      else
        Margin := (ClientSize.Y - TotalSize.Y + 1) div 2;
    end;
  end
  else
  begin
    if Spacing = -1 then
    begin
      TotalSize := Point(ClientSize.X - (Margin + GlyphSize.X), ClientSize.Y -
        (Margin + GlyphSize.Y));
      if Layout in [blGlyphLeft, blGlyphRight] then
        Spacing := (TotalSize.X - TextSize.X) div 2
      else
        Spacing := (TotalSize.Y - TextSize.Y) div 2;
    end;
  end;
    
  case Layout of
    blGlyphLeft:
      begin
        GlyphPos.X := Margin;
        TextPos.X := GlyphPos.X + GlyphSize.X + Spacing;
      end;
    blGlyphRight:
      begin
        GlyphPos.X := ClientSize.X - Margin - GlyphSize.X;
        TextPos.X := GlyphPos.X - Spacing - TextSize.X;
      end;
    blGlyphTop:
      begin
        GlyphPos.Y := Margin;
        TextPos.Y := GlyphPos.Y + GlyphSize.Y + Spacing;
      end;
    blGlyphBottom:
      begin
        GlyphPos.Y := ClientSize.Y - Margin - GlyphSize.Y;
        TextPos.Y := GlyphPos.Y - Spacing - TextSize.Y;
      end;
  end;

  { fixup the result variables }
  with GlyphPos do
  begin
    Inc(X, Client.Left + Offset.X);
    Inc(Y, Client.Top + Offset.Y);
  end;
  OffsetRect(TextBounds, TextPos.X + Client.Left + Offset.X,
    TextPos.Y + Client.Top + Offset.X);
end;
    
function TButtonGlyph.Draw(Canvas: TCanvas; const Client: TRect;
  const Offset: TPoint; const Caption: string; Layout: TButtonLayout;
  Margin, Spacing: Integer; State: TButtonState; Transparent: Boolean;
  BiDiFlags: LongInt): TRect;
var
  GlyphPos: TPoint;
begin
  CalcButtonLayout(Canvas, Client, Offset, Caption, Layout, Margin, Spacing,
    GlyphPos, Result, BiDiFlags);
  DrawButtonGlyph(Canvas, GlyphPos, State, Transparent);
  DrawButtonText(Canvas, Caption, Result, State, BiDiFlags);
end;

{ TExtBtn }
constructor TExtBtn.Create(AOwner: TComponent);
begin
  FGlyph := TButtonGlyph.Create;
  TButtonGlyph(FGlyph).OnChange := GlyphChanged;
  FExtGlyph := TButtonGlyph.Create;
  TButtonGlyph(FExtGlyph).Glyph.LoadFromResourceName(HInstance, 'CLOSE');
  inherited Create(AOwner);
  SetBounds(0, 0, 100, 22);
  ControlStyle := [csSetCaption, csCaptureMouse, csDoubleClicks];
  ParentFont := True;
  FSpacing := 4;
  FMargin := -1;
  FExtMargin := -1;
  FLayout := blGlyphLeft;
  FFlat := true;
  FTransparent := true;
  FExtWidth := 14;
  FExtButton := false;
  FBevelShow := true;
  FHotTrack := false;
  FThinBorder := false;
  Align := alNone;
  Inc(ButtonCount);

  FHotColor 	:= clBlue;
  FNormalColor 	:= clBlack;
  FNeedManualMouseUp := false;
  FDownColor := clBtnHighlight;
  FBtnColor := clBtnFace;
  Color := clBtnFace;

  TButtonGlyph(FGlyph).FBtnColor := FBtnColor;
  TButtonGlyph(FExtGlyph).FBtnColor := FBtnColor;
  Invalidate;
end;

destructor TExtBtn.Destroy;
begin
  Dec(ButtonCount);
  inherited Destroy;
  TButtonGlyph(FGlyph).Free;
  TButtonGlyph(FExtGlyph).Free;
end;

procedure TExtBtn.Paint;
const
  DownStyles: array[Boolean] of Integer = (BDR_RAISEDINNER, BDR_SUNKENOUTER);
  FillStyles: array[Boolean] of Integer = (BF_MIDDLE, 0);
var
  PaintRect: TRect;
  DrawFlags: Integer;
  Offset: TPoint;
begin
 	if not Enabled then begin
		FState := bsDisabled;
    	FDragging := False;
  	end else if FState = bsDisabled then
    	if FDown and (GroupIndex <> 0) then
      		FState := bsExclusive
    	else
      		FState := bsUp;
    Canvas.Font := Self.Font;

  	if(FExtButton)then	PaintRect := Rect(0, 0, Width-FExtWidth, Height)
	else   				PaintRect := Rect(0, 0, Width, Height);

	if not FFlat then begin
    	DrawFlags := DFCS_BUTTONPUSH or DFCS_ADJUSTRECT;
    	if FState in [bsDown, bsExclusive] then
      		DrawFlags := DrawFlags or DFCS_PUSHED;
    	DrawFrameControl(Canvas.Handle, PaintRect, DFC_BUTTON, DrawFlags);
  	end else begin
    	if ((((FState in [bsDown, bsExclusive]) or
      		(FMouseInControl and (FState <> bsDisabled)))or
      		(csDesigning in ComponentState)) and (FBevelShow or (FHotTrack and (FState in [bsExclusive, bsDown]))) or (FThinBorder))  then
	    DrawEdge(Canvas.Handle, PaintRect, DownStyles[FState in [bsDown, bsExclusive]], FillStyles[Transparent] or BF_RECT)
    	else if not Transparent then begin
			Canvas.Brush.Color := FBtnColor;
      		Canvas.FillRect(PaintRect);
    	end;
    	InflateRect(PaintRect, -1, -1);
  	end;
  	if FState in [bsDown, bsExclusive] then  begin
    	if ( not FTransparent and (FState in [bsExclusive]) and (not FFlat or not FMouseInControl or FHotTrack)) then begin
      		Canvas.Brush.Bitmap := AllocPatternBitmap(FBtnColor, FDownColor);
        	if (FBevelShow or FHotTrack) then Canvas.FillRect(PaintRect);
    	end;
    	if (FBevelShow or FHotTrack) then begin
      		Offset.X := 1;
      		Offset.Y := 1;
    	end else begin
      		Offset.X := 0;
      		Offset.Y := 0;
    	end;
  	end else begin
    	Offset.X := 0;
    	Offset.Y := 0;
	  	if (not FTransparent) then  begin
		    Canvas.Brush.Color := FBtnColor;
			Canvas.FillRect(PaintRect);
	    end;
  	end;
//  if ((FMouseInControl) and not(csDesigning in ComponentState)) then
//    TButtonGlyph(FGlyph).Draw(Canvas, PaintRect, Offset, Caption, FLayout, FMargin,
//      FSpacing, bsDown, Transparent, DrawTextBiDiModeFlags(0))
//  else
  	if (FMouseInControl and FHotTrack) then Font.Color := FHotColor
  	else									Font.Color := FNormalColor;

  	TButtonGlyph(FGlyph).Draw(Canvas, PaintRect, Offset, Caption, FLayout, FMargin, FSpacing, FState, Transparent, DrawTextBiDiModeFlags(0));

	if(FExtButton and ((FState in [bsDown, bsExclusive]) or FMouseInControl or (csDesigning in ComponentState) or not FFlat or (FThinBorder and FFlat)))then begin
		PaintRect := Rect(Width-FExtWidth, 0, Width, Height);
    	if not FFlat then begin
      		DrawFlags := DFCS_BUTTONPUSH or DFCS_ADJUSTRECT;
      		if FExtState in [bsDown, bsExclusive] then
        		DrawFlags := DrawFlags or DFCS_PUSHED;
      		DrawFrameControl(Canvas.Handle, PaintRect, DFC_BUTTON, DrawFlags);
    	end else begin
      		if (((FExtState in [bsDown, bsExclusive]) or (FMouseInControl and (FExtState <> bsDisabled)) or
        		(csDesigning in ComponentState)) and (FBevelShow or (FHotTrack and (FState in [bsExclusive, bsDown]))) or FThinBorder)  then
        		DrawEdge(Canvas.Handle, PaintRect, DownStyles[FExtState in [bsDown, bsExclusive]],
          	FillStyles[ExtTransparent] or BF_RECT)
      	else if not ExtTransparent then begin
			Canvas.Brush.Color := FBtnColor;
        	Canvas.FillRect(PaintRect);
      	end;
      	InflateRect(PaintRect, -1, -1);
    end;
    if FExtState in [bsDown, bsExclusive] then begin
      	if (FExtState = bsExclusive) and (not FFlat or not FMouseInControl) then begin
        	Canvas.Brush.Bitmap := AllocPatternBitmap(FBtnColor, clBtnHighlight);
        	Canvas.FillRect(PaintRect);
      	end;
      	Offset.X := 1;
      	Offset.Y := 1;
    end else begin
      	Offset.X := 0;
      	Offset.Y := 0;
	  	if (not FExtTransparent) then  begin
		    Canvas.Brush.Color := FBtnColor;
			Canvas.FillRect(PaintRect);
	    end;
    end;
    TButtonGlyph(FExtGlyph).Draw(Canvas, PaintRect, Offset, '', FLayout, FExtMargin, 0, FExtState, ExtTransparent, DrawTextBiDiModeFlags(0));
  end;
end;

procedure TExtBtn.UpdateMouseInControl;
begin
  if (FMouseInControl) then
  begin
    UpdateTracking;
  end;
  if (FNeedManualMouseUp and FDragging) then
  begin 
	MouseUp(mbLeft,[],0,0);
    FNeedManualMouseUp:=false;
  end;
end;

procedure TExtBtn.UpdateTracking;
var
  P: TPoint;
begin
  if FFlat then
  begin
    if Enabled then
    begin
      GetCursorPos(P);
      FMouseInControl := not (FindDragTarget(P, True) = Self);
      if FMouseInControl then
        Perform(CM_MOUSELEAVE, 0, 0)
      else
        Perform(CM_MOUSEENTER, 0, 0);
    end;
  end;
end;

procedure TExtBtn.Loaded;
var
  State: TButtonState;
begin
  inherited Loaded;
  if Enabled then
    State := bsUp
  else
    State := bsDisabled;
  TButtonGlyph(FGlyph).CreateButtonGlyph(State);
  TButtonGlyph(FExtGlyph).CreateButtonGlyph(State);
end;

procedure TExtBtn.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if (FNeedManualMouseUp) then
  begin
  	FNeedManualMouseUp := false;
  	exit;
  end;
  if (Button = mbLeft) and Enabled then
  begin
    if FExtButton and (X > (ClientWidth - FExtWidth))then
    begin
      FExtState := bsDown;
      Invalidate;
    end
    else
    if not FDown then
    begin
      FState := bsDown;
      Invalidate;
    end;
    if not (FGroupIndex = 0) and not (FDown) then
    begin
      FState := bsDown;
      Invalidate;
    end;
    FDragging := True;
  end;
end;

procedure TExtBtn.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  NewState: TButtonState;
begin
  inherited MouseMove(Shift, X, Y);
  if FDragging then
  begin
    if not FDown then NewState := bsUp
    else NewState := bsExclusive;
    if (X >= 0) and (X < ClientWidth) and (Y >= 0) and (Y <= ClientHeight) then
        if FDown then NewState := bsExclusive else NewState := bsDown;
    if NewState <> FState then
    begin
      FState := NewState;
      Invalidate;
    end;
  end
  else if not FMouseInControl then
    UpdateTracking;
end;

procedure TExtBtn.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  DoClick: Boolean;
  DoExtClick: Boolean;
begin
  inherited MouseUp(Button, Shift, X, Y);
  if (FExtState in [bsDown]) then
  begin
    fExtState := bsUp;
    Invalidate;
  end;
  if FDragging then
  begin
    FDragging := False;
    if FExtButton then
    begin
      DoClick := (X >= 0) and (X < (ClientWidth - FExtWidth)) and (Y >= 0) and (Y <= ClientHeight);
      DoExtClick := (X > (ClientWidth - FExtWidth)) and (X<ClientWidth) and (Y >= 0) and (Y <= ClientHeight);
    end
    else
    begin
      DoClick := (X >= 0) and (X < ClientWidth) and (Y >= 0) and (Y <= ClientHeight);
      DoExtClick := false;
    end;

    if FGroupIndex = 0 then
    begin
      { Redraw face in-case mouse is captured }
      FState := bsUp;
      FExtState := bsUp;
      FMouseInControl := False;
      if (DoClick or DoExtClick) and not (FState in [bsExclusive, bsDown]) then
        Invalidate;
    end
    else
      if DoClick or DoExtClick then
      begin
        if DoClick then SetDown(not FDown)
        else
          if (DoExtClick and not FDown) then SetDown(not FDown);
        if FDown then Repaint;
      end
      else
      begin
        if FDown then FState := bsExclusive;
        Repaint;
      end;
    if DoClick then Click;
    if DoExtClick then
    begin
      if Assigned(FOnExtBtnClick) then FOnExtBtnClick(Self);
    end;

    UpdateTracking;
  end;
end;

procedure TExtBtn.MouseManualUp;
begin
  FNeedManualMouseUp := true;
end;

procedure TExtBtn.Click;
begin
  inherited Click;
end;

function TExtBtn.GetPalette: HPALETTE;
begin
  Result := Glyph.Palette;
end;

function TExtBtn.GetGlyph: TBitmap;
begin
  Result := TButtonGlyph(FGlyph).Glyph;
end;

procedure TExtBtn.SetGlyph(Value: TBitmap);
begin
  TButtonGlyph(FGlyph).Glyph := Value;
  Invalidate;
end;

function TExtBtn.GetNumGlyphs: TNumGlyphs;
begin
  Result := TButtonGlyph(FGlyph).NumGlyphs;
end;

procedure TExtBtn.SetNumGlyphs(Value: TNumGlyphs);
begin
  if Value < 0 then Value := 1
  else if Value > 4 then Value := 4;
  if Value <> TButtonGlyph(FGlyph).NumGlyphs then
  begin
    TButtonGlyph(FGlyph).NumGlyphs := Value;
    Invalidate;
  end;
end;

procedure TExtBtn.GlyphChanged(Sender: TObject);
begin
  Invalidate;
end;

procedure TExtBtn.UpdateExclusive;
var
  Msg: TMessage;
begin
  if (FGroupIndex <> 0) and (Parent <> nil) then
  begin
    Msg.Msg := CM_BUTTONPRESSED;
    Msg.WParam := FGroupIndex;
    Msg.LParam := Longint(Self);
    Msg.Result := 0;
    Parent.Broadcast(Msg);
  end;
end;

procedure TExtBtn.SetDown(Value: Boolean);
begin
  if FGroupIndex = 0 then Value := False;
  if Value <> FDown then
  begin
    if FDown and (not FAllowAllUp) then Exit;
    FDown := Value;
    if Value then
    begin
      if FState = bsUp then Invalidate;
      FState := bsExclusive
    end
    else
    begin
      FState := bsUp;
      Repaint;
    end;
    if Value then UpdateExclusive;
  end;
end;

procedure TExtBtn.SetFlat(Value: Boolean);
begin
  if Value <> FFlat then
  begin
    FFlat := Value;
    Invalidate;
  end;
end;

procedure TExtBtn.SetGroupIndex(Value: Integer);
begin
  if FGroupIndex <> Value then
  begin
    FGroupIndex := Value;
    UpdateExclusive;
  end;
end;

procedure TExtBtn.SetLayout(Value: TButtonLayout);
begin
  if FLayout <> Value then
  begin
    FLayout := Value;
    Invalidate;
  end;
end;

procedure TExtBtn.SetMargin(Value: Integer);
begin
  if (Value <> FMargin) and (Value >= -1) then
  begin
    FMargin := Value;
    Invalidate;
  end;
end;

procedure TExtBtn.SetExtMargin(Value: Integer);
begin
  if (Value <> FExtMargin) and (Value >= -1) then
  begin
    FExtMargin := Value;
    Invalidate;
  end;
end;

procedure TExtBtn.SetSpacing(Value: Integer);
begin
  if Value <> FSpacing then
  begin
    FSpacing := Value;
    Invalidate;
  end;
end;

procedure TExtBtn.SetTransparent(Value: Boolean);
begin
  if Value <> FTransparent then
  begin
    FTransparent := Value;
    if Value then
      ControlStyle := ControlStyle - [csOpaque] else
      ControlStyle := ControlStyle + [csOpaque];
    Invalidate;
  end;
end;

procedure TExtBtn.SetAllowAllUp(Value: Boolean);
begin
  if FAllowAllUp <> Value then
  begin
    FAllowAllUp := Value;
    UpdateExclusive;
  end;
end;

procedure TExtBtn.WMLButtonDblClk(var Message: TWMLButtonDown);
begin
  inherited;
  if FDown then DblClick;
end;

procedure TExtBtn.CMEnabledChanged(var Message: TMessage);
const
  NewState: array[Boolean] of TButtonState = (bsDisabled, bsUp);
begin
  TButtonGlyph(FGlyph).CreateButtonGlyph(NewState[Enabled]);
  UpdateTracking;
  Repaint;
end;

procedure TExtBtn.CMButtonPressed(var Message: TMessage);
var
  Sender: TExtBtn;
begin
  if Message.WParam = FGroupIndex then
  begin
    Sender := TExtBtn(Message.LParam);
    if Sender <> Self then
    begin
      if Sender.Down and FDown then
      begin
        FDown := False;
        FState := bsUp;
        Invalidate;
      end;
      FAllowAllUp := Sender.AllowAllUp;
    end;
  end;
end;

procedure TExtBtn.CMDialogChar(var Message: TCMDialogChar);
begin
  with Message do
    if IsAccel(CharCode, Caption) and Enabled and Visible and
      (Parent <> nil) and Parent.Showing then
    begin
      Click;
      Result := 1;
    end else
      inherited;
end;

procedure TExtBtn.CMFontChanged(var Message: TMessage);
begin
  Invalidate;
end;

procedure TExtBtn.CMTextChanged(var Message: TMessage);
begin
  Invalidate;
end;

procedure TExtBtn.CMSysColorChange(var Message: TMessage);
begin
  with TButtonGlyph(FGlyph) do
  begin
    Invalidate;
    CreateButtonGlyph(FState);
  end;
end;

procedure TExtBtn.CMMouseEnter(var Message: TMessage);
begin
  inherited;
  { Don't draw a border if DragMode <> dmAutomatic since this button is meant to
    be used as a dock client. }
  if FFlat and not FMouseInControl and Enabled and (DragMode <> dmAutomatic)
    and (GetCapture = 0) then
  begin
    FMouseInControl := True;
    Repaint;
  end;
end;

procedure TExtBtn.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  if FFlat and FMouseInControl and Enabled and not FDragging then
  begin
    FMouseInControl := False;
    Invalidate;
  end;
end;

procedure TExtBtn.ActionChange(Sender: TObject; CheckDefaults: Boolean);
  procedure CopyImage(ImageList: TCustomImageList; Index: Integer);
  begin
    with Glyph do
    begin
      Width := ImageList.Width;
      Height := ImageList.Height;
      Canvas.Brush.Color := clFuchsia;//! for lack of a better color
      Canvas.FillRect(Rect(0,0, Width, Height));
      ImageList.Draw(Canvas, 0, 0, Index);
    end;
  end;

begin
  inherited ActionChange(Sender, CheckDefaults);
  if Sender is TCustomAction then
    with TCustomAction(Sender) do
    begin
      { Copy image from action's imagelist }
      if (Glyph.Empty) and (ActionList <> nil) and (ActionList.Images <> nil) and
        (ImageIndex >= 0) and (ImageIndex < ActionList.Images.Count) then
        CopyImage(ActionList.Images, ImageIndex);
    end;
end;

procedure TExtBtn.SetExtWidth(Value: Integer);
begin
  if (Value <> FExtWidth) and (Value >= -1) and (Value <= width div 2) then
  begin
    FExtWidth := Value;
    Invalidate;
  end;
end;

procedure TExtBtn.SetExtButton(Value: Boolean);
begin
  if Value <> FExtButton then
  begin
    FExtButton := Value;
    Invalidate;
  end;
end;

procedure TExtBtn.SetBevelShow(Value: Boolean);
begin
  if Value <> FBevelShow then
  begin
    FBevelShow := Value;
    Invalidate;
  end;
end;

procedure TExtBtn.SetBtnColor(Value: TColor);
begin
  if Value <> FBtnColor then
  begin
    FBtnColor := Value;
	TButtonGlyph(FGlyph).FBtnColor 		:= FBtnColor;
	TButtonGlyph(FExtGlyph).FBtnColor 	:= FBtnColor;
    Invalidate;
  end;
end;

procedure TExtBtn.SetNormalColor(Value: TColor);
begin
  if Value <> FNormalColor then
  begin
    FNormalColor 						:= Value;
    Font.Color							:= FNormalColor;
    Invalidate;
  end;
end;

procedure TExtBtn.SetDownColor(Value: TColor);
begin
  if Value <> FDownColor then
  begin
    FDownColor := Value;
    Invalidate;
  end;
end;

procedure TExtBtn.SetHotTrack(Value: Boolean);
begin
  if Value <> FHotTrack then
  begin
    FHotTrack := Value;
    Invalidate;
  end;
end;

procedure TExtBtn.SetExtTransparent(Value: Boolean);
begin
  if Value <> FExtTransparent then
  begin
    FExtTransparent := Value;
    Invalidate;
  end;
end;

procedure TExtBtn.SetThinBorder(Value: Boolean);
begin
  if Value <> FThinBorder then
  begin
    FThinBorder := Value;
    Invalidate;
  end;
end;

procedure TExtBtn.SetKind(Value: TKind);
begin
  if Value <> FKind then
  begin
//    TButtonGlyph(FExtGlyph).Glyph.ReleaseHandle;
    TButtonGlyph(FExtGlyph).Free;
    FExtGlyph := TButtonGlyph.Create;
    FKind := Value;
    case FKind of
      knClose: TButtonGlyph(FExtGlyph).Glyph.LoadFromResourceName(HInstance, 'CLOSE');
      knMinimize: TButtonGlyph(FExtGlyph).Glyph.LoadFromResourceName(HInstance, 'MINIMIZED');
//      knNone: TButtonGlyph(FExtGlyph).Glyph.LoadFromResourceName(HInstance, 'MINIMIZED');
    end;
	TButtonGlyph(FGlyph).FBtnColor 		:= FBtnColor;
	TButtonGlyph(FExtGlyph).FBtnColor 	:= FBtnColor;
    Invalidate;
  end;
end;
end.

